class Error(Exception):
    pass


class FQDNMustNotBeEmptyError(Error, ValueError):
    def __init__(self):
        super().__init__(f"Error: FQDN must not be empty")


class MissingMetadataError(Error, ValueError):
    def __init__(self, src):
        super().__init__(f"missing metadata in template: {src}")


class MissingDestinationError(Error, ValueError):
    def __init__(self, src):
        super().__init__(f"missing `dest` in template metadata: {src}")


class InvalidDestinationError(Error, ValueError):
    def __init__(self, dest):
        super().__init__(f"invalid destination: {dest}")


class InfraDotPyNotFoundError(Error, ImportError):
    def __init__(self, modulename):
        super().__init__(f"{modulename}.py: file not found")


class HostOrServiceNotFoundError(Error, KeyError):
    def __init__(self, name):
        super().__init__(f"{name}: host or service not found")


class SwitchValueError(Error, ValueError):
    def __init__(self, path, value, options):
        options = ", ".join(options)
        super().__init__(
            f"Invalid value {value} for switch {path}. Options are: {options}"
        )


class SwitchCaseError(Error, ValueError):
    def __init__(self, path, options, cases):
        _options = ", ".join(options)
        _cases = ", ".join(cases.keys())

        super().__init__(
            f"{path}: cases are incompatible with the options."
            f" Options are: {_options}, Cases are: {_cases}"
        )


class SwitchError(Error, ValueError):
    def __init__(self, path, val, cases):
        _cases = ", ".join(cases.keys())
        super().__init__(
            f"{path}: switch value is set to {val},"
            f" but no case matched. Cases are: {_cases}"
        )


class DuplicateNameError(Error, ValueError):
    def __init__(self, name):
        super().__init__(f"Duplicate name: {name}")


class DuplicateIPError(Error, ValueError):
    def __init__(self, ip):
        super().__init__(f"Duplicate IP: {ip}")


class DuplicateFQDNError(Error, ValueError):
    def __init__(self, fqdn):
        super().__init__(f"Duplicate IP: {fqdn}")



from dataclasses import asdict, dataclass, field

from sh import Command, ssh
from typer import Typer

from infrable.meta import Meta
from infrable.utils import format_item


@dataclass(unsafe_hash=True, order=True, eq=True)
class Host:
    """A generic host in the infrastructure."""

    ip: str  # The only required field

    fqdn: str | None = None
    admin: str = "root"
    meta: Meta = field(default_factory=Meta)
    typer: Typer | None = None

    def remote(self, login=None) -> Command:
        login = login or self.admin
        address = f"{login}@{self.ip}"
        return ssh.bake(address, "-o", "StrictHostKeyChecking=no")

    def format(self, name: str, format: str) -> str:
        return format_item(format).render(name=name, **asdict(self))

    def __str__(self) -> str:
        host = self.fqdn or self.ip
        return f"{self.admin}@{host}"



__version__ = "0.1.15"


from pathlib import Path

INFRA_MODULE_NAME = "infra"
IS_PROJECT_DIR = Path(f"{INFRA_MODULE_NAME}.py").exists()

if IS_PROJECT_DIR:
    from infrable.distro.linux import LinuxLatestHost
    from infrable.distro.ubuntu import UbuntuLatestHost
    from infrable.host import Host
    from infrable.meta import Meta
    from infrable.readfile import readfile
    from infrable.service import Service
    from infrable.switch import Switch
    from infrable.utils import concurrent, concurrentcontext



from dataclasses import asdict, dataclass, field
from typing import Any

from typer import Typer

from infrable import errors, paths
from infrable.meta import Meta
from infrable.utils import format_item


@dataclass(unsafe_hash=True, order=True, eq=True)
class Switch:
    """A switch for the infrastructure."""

    options: set[str]
    init: str

    typer: Typer | None = None
    meta: Meta = field(default_factory=Meta)

    def __post_init__(self):
        self.path = paths.switches / "-".join(sorted(self.options))
        if not self.path.exists():
            self.set(self.init)

    def __call__(self, **cases: Any) -> Any | None:
        """Get the current value of the switch. returns None if not defined."""

        if not cases:
            cases = {opt: opt for opt in self.options}
        elif len(set(cases.keys()) - self.options) > 0:
            raise errors.SwitchCaseError(self.path, options=self.options, cases=cases)

        val = self.path.read_text().strip()
        return cases.get(val)

    def strict(self, **cases: Any) -> Any:
        """Strictly require all cases to be defined."""

        if not cases:
            cases = {opt: opt for opt in self.options}
        elif set(cases.keys()) != self.options:
            raise errors.SwitchCaseError(self.path, options=self.options, cases=cases)

        val = self.path.read_text().strip()
        if val in cases:
            return cases[val]

        raise errors.SwitchError(self.path, val=val, cases=cases)

    def set(self, value: str):
        if value in self.options:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self.path.write_text(value)
            return

        self.path.unlink(missing_ok=True)
        raise errors.SwitchValueError(self.path, value=value, options=self.options)

    def format(self, name: str, format: str) -> str:
        return format_item(format).render(name=name, **asdict(self))

    def __str__(self) -> str:
        return self.strict()



from pathlib import Path
from typing import Callable


def readfile(
    path: Path | str,
    *,
    init: Callable[[], str] = str,
    fmt: Callable[[str], str] | None = lambda s: s.strip(),
) -> str:
    """Get the contents of a file, creating it if it doesn't exist."""

    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        val = init()
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(val)
    else:
        val = path.read_text()
    if fmt:
        return fmt(val)
    return val



from pathlib import Path

INFRA_PY = '''
import typer
from infrable import Meta, Service, Switch
from infrable import UbuntuLatestHost as Host
from infrable import concurrent, files, paths, readfile
from lib.modules import mycloud

# To quickly find the configuration file template on GitHub
template_prefix = "https://github.com/username/repository/blob/main"


# Environments/ ----------------------------------------------------------------
dev = "dev"
beta = "beta"
prod = "prod"

environments = {dev, beta, prod}
env = Switch(environments, init=dev)
current_env = env()
# /Environments ----------------------------------------------------------------


# Clouds/ ----------------------------------------------------------------------
cloud = mycloud.MyCloud(secret_api_key=readfile("secrets/mycloud/secret_api_key"))
cloud.typer = mycloud.workflows
# /Clouds ----------------------------------------------------------------------


# Hosts/ -----------------------------------------------------------------------
dev_host = Host(fqdn="dev.example.com", ip="127.0.0.1")
beta_host = Host(fqdn="beta.example.com", ip="127.0.0.2")
prod_host = Host(fqdn="prod.example.com", ip="127.0.0.3")

managed_hosts = env(
    dev=[dev_host],
    beta=[beta_host],
    prod=[prod_host],
)
# /Hosts -----------------------------------------------------------------------


# Services/ --------------------------------------------------------------------
web = Service(
    host=env.strict(dev=dev_host, beta=beta_host, prod=prod_host),
    meta=Meta(secret_key=readfile("secrets/web/secret_key")),
    port=8080,
)

nginx = Service(port=80, host=web.host)
# /Services --------------------------------------------------------------------


# Tasks/ -----------------------------------------------------------------------
nginx.typer = typer.Typer(help="Nginx specific tasks.")
@nginx.typer.command(name="reload")
def reload_nginx():
    """[TASK] Reload nginx: infrable nginx reload"""

    assert nginx.host, "Service must have a host to reload"
    nginx.host.remote().sudo.nginx("-t")
    nginx.host.remote().sudo.systemctl.reload.nginx()
# Tasks/ -----------------------------------------------------------------------

# Workflows/ -----------------------------------------------------------------------
deploy = typer.Typer(help="Deployment workflows.")
@deploy.command(name="nginx")
def deploy_nginx():
    """[WORKFLOW] Deploy nginx files: infrable deploy nginx"""

    files.deploy(paths.templates / "nginx")
    cmd = "sudo nginx -t && sudo systemctl reload nginx && echo success || echo failed"
    fn = lambda host: (host, host.remote().sudo(cmd))
    for host, result in concurrent(fn, files.affected_hosts()):
        print(f"{host}: {result}")
# /Workflows -----------------------------------------------------------------------
'''.strip()

NGINX_PROXY_PARAMS_TEMPLATE = """
# vim: syn=nginx

# ---
# src: {{ template_prefix }}/{{ _template.src }}
# dest:
# {% for host in managed_hosts %}
#   - loc: {{ host }}:/etc/nginx/proxy_params
# {% endfor %}
# chmod: 644
# chown: root:root
# ---
proxy_set_header Host $http_host;
proxy_set_header X-Real-IP $remote_addr;
proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
proxy_set_header X-Forwarded-Proto $scheme;
""".strip()

NGINX_WEB_TEMPLATE = """
# vim: syn=nginx

# ---
# src: {{ template_prefix }}/{{ _template.src }}
# dest: {{ nginx.host }}:/etc/nginx/sites-enabled/web
# chmod: 644
# chown: root:root
# ---

server {
    listen {{ nginx.port }};
    listen [::]:{{ nginx.port }};

    server_name {{ nginx.host.fqdn }} www.{{ nginx.host.fqdn }};

    location / {
        proxy_pass http://127.0.0.1:{{ web.port }};
        include proxy_params;
    }
}
""".strip()

MYCLOUD_MODULE = '''
from dataclasses import dataclass
from typer import Typer
from infrable import infra
from infrable.distro.ubuntu import UbuntuLatestHost


@dataclass
class MyCloud:
    """MyCloud Python library."""

    secret_api_key: str
    typer: Typer | None = None

    def provision_ubuntu_host(self, fqdn: str):
        ip = self.api.create_ubuntu_host(fqdn)
        return MyCloudUbuntuHost(fqdn=fqdn, ip=ip)

@dataclass
class MyCloudUbuntuHost(UbuntuLatestHost):
    """MyCloud's customized Ubuntu server."""

    def setup(self, as_root: bool = False):
        super().setup(as_root=as_root)
        self.install_mycloud_agent()

    def install_mycloud_agent(self):
        raise NotImplementedError

workflows = Typer()

@workflows.command()
def provision_ubuntu_host(fqdn: str, setup: bool = True):
    """[WORKFLOW] Provision Ubuntu host."""

    # Get the MyCloud instance from infra.py
    cloud = next(iter(infra.item_types[MyCloud].values()))

    # Provision the host
    host = cloud.provision_ubuntu_host(fqdn)
    if setup:
        host.setup(as_root=True)

    name = fqdn.split(".")[0]
    print("Add the host to the infra.py file.")
    print(f"{name} = {repr(host)}")
'''.strip()

FILES = {
    "infra.py": INFRA_PY,
    "templates/nginx/proxy_params.j2": NGINX_PROXY_PARAMS_TEMPLATE,
    "templates/nginx/web.j2": NGINX_WEB_TEMPLATE,
    "lib/modules/mycloud.py": MYCLOUD_MODULE,
}


def init():
    for filename, content in FILES.items():
        path = Path(filename)
        if path.exists() and path.read_text().strip():
            print(f"Skipping {path}, Exists.")
            continue
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(content)
        print(f"Created {path}.")



from dataclasses import dataclass, field
from shlex import quote as q
from typing import Optional

from infrable import errors
from infrable.host import Host


@dataclass(unsafe_hash=True, order=True, eq=True)
class LinuxLatestHost(Host):
    """A latest version of generic Linux host."""

    authorized_keys: str = ""
    users: dict[str, dict[str, str]] = field(default_factory=dict)

    def setup(self, as_root: bool = False):
        self.ensure_admin(as_root=as_root)
        self.disable_root_login()
        self.ensure_hostname()
        self.ensure_users()

    def ensure_hostname(self, fqdn: str | None = None):
        fqdn = fqdn or self.fqdn
        if not fqdn:
            raise errors.FQDNMustNotBeEmptyError()
        self.remote().sudo.tee("/etc/hostname", _in=fqdn.split(".", 1)[0], _fg=True)

    def ensure_admin(
        self,
        username: str | None = None,
        authorized_keys: str | None = None,
        as_root: bool = False,
    ):
        username = username or self.admin
        self.ensure_user(username, as_root=as_root)
        self.ensure_authorized_keys(username, keys=authorized_keys, as_root=as_root)
        self.ensure_sudoer(username, as_root=as_root)

    def ensure_sudoer(
        self,
        username: Optional[str] = None,
        access: str = "ALL=(ALL) NOPASSWD:ALL",
        as_root: bool = False,
    ):
        conn = self.remote("root") if as_root else self.remote().sudo
        username = username or self.admin
        if not conn.ls(q(f"/etc/sudoers.d/{username}"), _ok_code=[0, 2]):
            conn.tee(
                q(f"/etc/sudoers.d/{username}"),
                _in=f"{username} {access}",
            )

            conn.chmod("0440", q(f"/etc/sudoers.d/{username}"))
        self.remote(username).sudo.echo("sudo setup successful", _fg=True)

    def ensure_user(
        self, username: str, shell: str = "/bin/bash", as_root: bool = False
    ):
        conn = self.remote("root") if as_root else self.remote().sudo
        if not conn.id(q(username), _ok_code=[0, 1]):
            conn.useradd("-m", "-s", q(shell), q(username), _fg=True)

    def ensure_users(self, users: dict[str, dict[str, str]] | None = None):
        for username, config in (users or self.users).items():
            self.ensure_user(username, **config, as_root=False)

    def ensure_authorized_keys(
        self, username: str, keys: str | None, as_root: bool = False
    ):
        conn = self.remote("root") if as_root else self.remote().sudo
        sshdir = f"/home/{username}/.ssh"
        dest = f"/{sshdir}/authorized_keys"
        conn.mkdir("-pv", q(sshdir), _fg=True)
        conn.tee(q(dest), _in=keys or self.authorized_keys)
        conn.chown("-R", q(f"{username}:{username}"), sshdir, _fg=True)
        conn.chmod("0600", dest, _fg=True)
        self.remote(login=username).echo(f"{username}: ssh setup successful", _fg=True)

    def disable_root_login(self):
        sudo_conn = self.remote().sudo
        if sudo_conn.grep(
            q("^PermitRootLogin\\syes"), "/etc/ssh/sshd_config", _ok_code=[0, 1]
        ):
            sudo_conn.sed("-i", "/^PermitRootLogin/s/yes/no/", "/etc/ssh/sshd_config")
            sudo_conn.systemctl.reload.ssh()



from pathlib import Path

from sh import infrable

from infrable import __version__, init

data = Path("tests/data")


def setup_module():
    infrable.switch.env.dev()
    init.init()


def test_switch():
    assert infrable.switches(_tty_out=False).strip() == "env = dev"
    infrable.switch.env.beta()
    assert infrable.switches(_tty_out=False).strip() == "env = beta"
    infrable.switch.env.prod()
    assert infrable.switches(_tty_out=False).strip() == "env = prod"



import random
import shutil
import subprocess as sp
import time
from datetime import datetime
from difflib import unified_diff
from pathlib import Path
from shlex import quote as q
from typing import IO, Any, Iterable

import typer
from click import Choice
from jinja2 import Environment, FileSystemLoader
from tqdm import tqdm

from infrable import errors, infra, paths, utils
from infrable.host import Host
from infrable.template import Metadata, render


def affected_hosts(only: Iterable[str] | None = None) -> list[Host]:
    """List the affected hosts as per the files diff."""

    _only = list(infra.filtered_hosts(only=only))
    only_hosts = set(h.ip for h in _only) | set(h.fqdn for h in _only if h.fqdn)

    hosts = {}
    for new in paths.files.glob("**/*.new"):
        old = _old_path(new)
        if not _diff(new=new, old=old):
            continue
        remote, _ = str(new.relative_to(paths.files)).removesuffix(".new").split("/", 1)
        host = remote.split("@", 1)[-1]
        if only and host not in only_hosts:
            continue
        if host not in hosts:
            hosts[host] = infra.get_host(host)
    return list(hosts.values())


def gen(path: Path | None = None, only: Iterable[str] | None = None):
    """Generate files by rendering the templates with context from infra."""

    if paths.files.exists():
        shutil.rmtree(str(paths.files))

    path = path or paths.templates
    templates = Environment(loader=FileSystemLoader(path))

    _only = list(infra.filtered_hosts(only=only))
    only_hosts = set(h.ip for h in _only) | set(h.fqdn for h in _only if h.fqdn)
    for t in templates.list_templates():
        # Render the template metadata
        src = path / t
        typer.secho(f"╭ {src}", bold=True)

        meta = Metadata.parse(src, infra.items)

        if meta.skip:
            print("╰ skip: true")
            print()
            continue

        for i, dest in enumerate(meta.dest):
            if i + 1 == len(meta.dest):
                print("╰─ ", end="")
            else:
                print("├─ ", end="")

            # Render the template for each destination

            remote, loc = dest.loc.split(":", 1)
            host = remote.split("@", 1)[-1]
            if not host or host == "None" or (only and host not in only_hosts):
                print(f"skipping {dest.loc}")
                continue
            outpath = paths.files / remote / (loc.removeprefix("/") + ".new")
            outpath.parent.mkdir(exist_ok=True, parents=True)
            rendered = render(src, dest.loc, infra.items, **dest.ctx)
            outpath.write_text(rendered)
            print(outpath)
        print()


def pull(workers: int | None = None):
    """For each generated file pull the current version from the server."""

    if not paths.files.exists():
        return

    for old in paths.files.glob("**/*.old"):
        old.unlink()

    new_files = list(paths.files.glob("**/*.new"))
    with utils.concurrentcontext(_pull, new_files, workers=workers) as results:
        for new, dest in tqdm(results, total=len(new_files)):
            typer.secho(f"╭ {new}", bold=True)
            print(f"╰ {dest}")
            print()


def backup():
    """Backup the files locally."""

    if not paths.files.exists():
        return

    dirname = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
    backupdir = paths.backups / dirname
    while backupdir.exists():
        time.sleep(1)
        dirname = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        backupdir = paths.backups / dirname
    shutil.copytree(str(paths.files), str(backupdir))
    print(f"backup: {backupdir}")


def diff():
    """Diff the generated files with the pulled versions."""

    for new in paths.files.glob("**/*.new"):
        old = _old_path(new)
        if d := _diff(new=new, old=old):
            _print_diff(d)
            print()


def push(yes: bool = False):
    """Diff and push the generated files to the remote server."""

    for new in paths.files.glob("**/*.new"):
        remote, loc = (
            str(new.relative_to(paths.files)).removesuffix(".new").split("/", 1)
        )

        old = _old_path(new)
        remote_dest = f"{remote}:/{loc}"

        if d := _diff(new=new, old=old):
            _print_diff(d)
            if not yes:
                ans = typer.prompt(
                    f"Push?", type=Choice(["y", "n", "all"]), default="y"
                )
                match ans.lower():
                    case "n":
                        new.unlink()
                        old.unlink()
                        continue
                    case "all":
                        yes = True

            meta = Metadata.parse(new)
            dest = next(d for d in meta.dest if d.loc == remote_dest)
            chown = dest.chown or meta.chown
            chmod = dest.chmod or meta.chmod
            dest = f"/{loc}"
            name = remote.split("@", 1)[-1]
            host = infra.get_host(name)
            if not host:
                raise errors.InvalidDestinationError(dest)
            addr = f"{host.admin}@{host.ip}"
            _sudo_push_file(addr, new, dest, chown=chown, chmod=chmod)
        else:
            print(f"skipping {remote_dest}")
            new.unlink()
            old.unlink()
        print()


def revert(path: Path | None = None):
    """Revert files from backup."""

    if path is None:
        path = max(paths.backups.glob("*"))

    print(f"reverting from: {path}")

    for bkold in path.glob("**/*.old"):
        bknew = Path(str(bkold).removesuffix(".old") + ".new")

        old = paths.files / bkold.relative_to(path)
        new = paths.files / bknew.relative_to(path)

        typer.secho(f"╭ {new}", bold=True)
        shutil.copy(str(bkold), str(new))

        print(f"╰ {old}")
        shutil.copy(str(bknew), str(old))


def deploy(
    path: Path | None = None, only: Iterable[str] | None = None, yes: bool = False
):
    """[WORKFLOW] generate, pull, backup, compare and push the generated files to the remote server."""

    if path or yes or typer.confirm("Generate fresh files?", default=True, err=True):
        gen(path, only=only)
        print()
        pull()
        print()
        backup()
    elif typer.confirm("Pull files from remote?", default=True, err=True):
        pull()
        print()
        backup()
    elif typer.confirm("Backup files locally?", default=True, err=True):
        backup()

    print()
    push(yes=yes)


def recover(path: Path | None = None):
    """[WORKFLOW] revert and push files from backup."""

    revert(path)
    print()
    push()


def _pull(new):
    dest = str(new).removesuffix(".new") + ".old"
    remote, loc = str(new.relative_to(paths.files)).removesuffix(".new").split("/", 1)
    src = f"/{loc}"
    name = remote.split("@", 1)[-1]
    host = infra.get_host(name)
    if not host:
        raise errors.InvalidDestinationError(dest=dest)
    addr = f"{host.admin}@{host.ip}"
    _sudo_pull_file(addr, src, dest)
    return new, dest


def _old_path(new: Path) -> Path:
    return Path(str(new).removesuffix(".new") + ".old")


def _diff(new: Path, old: Path) -> list[str]:
    new_lines = new.read_text().splitlines()
    old_lines = Path(old).read_text().splitlines()
    return list(unified_diff(old_lines, new_lines, fromfile=str(old), tofile=str(new)))


def _print_diff(diff: list[str]):
    for line in diff:
        if line.startswith("---"):
            typer.secho(line, fg=typer.colors.RED, bold=True)
        elif line.startswith("+++"):
            typer.secho(line, fg=typer.colors.GREEN, bold=True)
        elif line.startswith("-"):
            typer.secho(line, fg=typer.colors.RED)
        elif line.startswith("+"):
            typer.secho(line, fg=typer.colors.GREEN)

        elif line.startswith("@"):
            typer.secho(line, dim=True)
        else:
            print(line)


def _sudo_exec(
    address: str,
    command: list[str] | str,
    options: list[str] | None = None,
    shell: str | None = None,
    stdout: int | IO[Any] = sp.PIPE,
    stderr: int | IO[Any] = sp.PIPE,
    stdin: int | IO[Any] = sp.PIPE,
) -> sp.CompletedProcess[bytes]:
    if isinstance(command, list):
        command = sp.list2cmdline(command)
    if shell:
        command = sp.list2cmdline([shell, "-c", q(command)])
    options = options or []
    return sp.run(
        [
            "ssh",
            "-o",
            "StrictHostKeyChecking=no",
            *options,
            address,
            command,
        ],
        stdout=stdout,
        stderr=stderr,
        stdin=stdin,
    )


def _sudo_pull_file(address: str, remote_src: str, local_dest: str):
    with Path(local_dest).open("w") as f:
        _sudo_exec(
            address,
            f"sudo cat -- {q(remote_src)} || true",
            stdout=f,
            stderr=sp.PIPE,
            stdin=sp.PIPE,
        ).check_returncode()


def _sudo_push_file(
    address: str,
    local_src,
    remote_dest,
    chmod: str | None = None,
    chown: str | None = None,
):
    ts = time.time()
    rand = random.randint(0, 9999)
    tmpname = f"/tmp/infrable-push-{ts}.{rand}"
    parent = str(Path(remote_dest).parent)
    cmd = f"sudo tee -- {q(tmpname)}"  # as root for security

    # Change permissions and ownership before moving to final destination.
    # Else, you might end up getting locked from the server.
    if chmod:
        cmd = f"{cmd} && sudo chmod -- {q(chmod)} {q(tmpname)}"

    if chown:
        cmd = f"{cmd} && sudo chown -- {q(chown)} {q(tmpname)}"

    cmd += f" && sudo mkdir -p -- {q(parent)}"
    cmd += f" && sudo mv -f -- {q(tmpname)} {q(remote_dest)}"

    with Path(local_src).open("r") as f:
        _sudo_exec(address, cmd, stdin=f).check_returncode()



from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Any, Callable, Generator, Iterable, Iterator, TypeVar

from jinja2 import Template

T = TypeVar("T")


def format_item(template: str) -> Template:
    return Template(
        template,
        variable_start_string="{",
        variable_end_string="}",
        autoescape=False,
        keep_trailing_newline=True,
        finalize=lambda x: x or "",
    )


@contextmanager
def concurrentcontext(
    function: Callable[[T], Any],
    generator: Iterable[T],
    *,
    workers: int | None = None,
) -> Generator[Iterator[Any], Any, None]:
    """With context, run a function on a batch of arguments concurrently."""

    with ThreadPoolExecutor(max_workers=workers) as executor:
        yield executor.map(function, generator)


def concurrent(
    function: Callable[[T], Any],
    generator: Iterable[T],
    *,
    workers: int | None = None,
) -> list:
    """Run a functions on a batch of arguments in concurrently."""

    with concurrentcontext(function, generator, workers=workers) as results:
        return list(results)



from dataclasses import dataclass, field
from shlex import quote as q

from infrable.distro.linux import LinuxLatestHost


@dataclass(unsafe_hash=True, order=True, eq=True)
class UbuntuLatestHost(LinuxLatestHost):
    """A latest version of Ubuntu."""

    packages: list[str] = field(default_factory=list)

    def setup(self, as_root: bool = False):
        super().setup(as_root=as_root)
        self.upgrade_packages()
        self.update_packages()
        self.ensure_apt_packages()

    def upgrade_packages(self):
        self.remote().sudo(
            "apt-get", "-o", "DPkg::Lock::Timeout=60", "upgrade", "-y", _fg=True
        )

    def update_packages(self):
        self.remote().sudo(
            "apt-get", "-o", "DPkg::Lock::Timeout=60", "update", "-y", _fg=True
        )

    def ensure_apt_packages(self, packages: list[str] | None = None):
        packages = packages or self.packages
        if packages:
            self.remote().sudo(
                "apt-get", "-o", "DPkg::Lock::Timeout=60", "update", "-y", _fg=True
            )
            package_batches = [
                self.packages[i : i + 10] for i in range(0, len(packages), 10)
            ]
            for batch in package_batches:
                self.remote().sudo(
                    "apt-get",
                    "-o",
                    "DPkg::Lock::Timeout=60",
                    "install",
                    "-y",
                    *[q(pkg) for pkg in batch],
                    _fg=True,
                )
            self.remote().sudo(
                "apt-get", "-o", "DPkg::Lock::Timeout=60", "autoremove", "-y", _fg=True
            )

    def disable_root_login(self):
        sudo_conn = self.remote().sudo
        if sudo_conn.grep(
            q("^PermitRootLogin\\syes"), "/etc/ssh/sshd_config", _ok_code=[0, 1]
        ):
            sudo_conn.sed("-i", "/^PermitRootLogin/s/yes/no/", "/etc/ssh/sshd_config")
            sudo_conn.systemctl.reload.ssh()



import typer
from click import Choice

from infrable import infra

app = typer.Typer(no_args_is_help=True)


for name, sw in infra.switches.items():
    help = f"Set the value of the {name} switch."

    def main(value: str = typer.Argument(..., click_type=Choice(list(sw.options)))):
        sw.set(value)

    app.command(name=name, help=help)(main)



import shutil
from pathlib import Path

from sh import infrable

from infrable import __version__, init, paths

data = Path("tests/data")


def gen_old_files(copy: bool):
    for new in paths.files.glob("**/*.new"):
        old = (paths.files / new.relative_to(paths.files)).with_suffix(".old")
        text = new.read_text() if copy else ""
        old.write_text(text)


def gen_test_data():
    infrable.files.gen(_out=data.joinpath("pre_files_gen"))
    infrable.files.gen(_out=data.joinpath("post_files_gen"))
    gen_old_files(copy=False)
    infrable.files.diff(_out=data.joinpath("files_diff"))


def setup_module():
    infrable.switch.env.dev()
    # gen_test_data()
    init.init()


def test_files_gen():
    shutil.rmtree(paths.files, ignore_errors=True)
    assert (
        infrable.files.gen(_tty_out=False) == data.joinpath("pre_files_gen").read_text()
    )
    assert (
        infrable.files.gen(_tty_out=False)
        == data.joinpath("post_files_gen").read_text()
    )


def test_files_diff():
    shutil.rmtree(paths.files, ignore_errors=True)
    infrable.files.gen()
    gen_old_files(copy=True)
    assert infrable.files.diff(_tty_out=False) == ""
    gen_old_files(copy=False)
    assert (
        infrable.files.diff(_tty_out=False) == data.joinpath("files_diff").read_text()
    )


def test_files_backup():
    infrable.files.gen()
    gen_old_files(copy=False)
    backups_count = len(list(paths.backups.glob("*")))
    out = infrable.files.backup(_tty_out=False)
    assert out.strip().startswith(f"backup: {paths.backups}/")
    assert len(list(paths.backups.glob("*"))) == backups_count + 1


def test_files_revert():
    infrable.files.gen()
    gen_old_files(copy=False)
    infrable.files.backup()  # <-- this fails for some reason

    news1, olds1 = [f.read_text() for f in paths.files.glob("**/*.new")], [
        f.read_text() for f in paths.files.glob("**/*.old")
    ]

    out = infrable.files.revert(_tty_out=False)
    assert out.strip().startswith(f"reverting from: {paths.backups}/")

    news2, olds2 = [f.read_text() for f in paths.files.glob("**/*.new")], [
        f.read_text() for f in paths.files.glob("**/*.old")
    ]

    assert news1 == olds2
    assert olds1 == news2


def test_files_affect_hosts():
    infrable.files.gen()
    gen_old_files(copy=True)
    assert infrable.files("affected-hosts", _tty_out=False) == ""
    gen_old_files(copy=False)
    assert (
        infrable.files("affected-hosts", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="dev_host", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="dev.example.com", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="127.0.0.1", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="managed_hosts", _tty_out=False).strip()
        == "root@dev.example.com"
    )
    assert (
        infrable.files("affected-hosts", only="prod_host", _tty_out=False).strip() == ""
    )



from dataclasses import asdict, dataclass, field

from typer import Typer

from infrable.host import Host
from infrable.meta import Meta
from infrable.utils import format_item


@dataclass(unsafe_hash=True, order=True, eq=True)
class Service:
    """A generic service, maybe running on a host."""

    host: Host | None = None
    port: int | None = None
    meta: Meta = field(default_factory=Meta)
    typer: Typer | None = None

    def format(self, name: str, format: str) -> str:
        return format_item(format).render(name=name, **asdict(self))

    def __str__(self) -> str:
        return f'{self.host or ""}:{self.port or ""}'



from pathlib import Path

from sh import infrable

from infrable import __version__, init

data = Path("tests/data")


def gen_test_data():
    infrable.switches(_out=data.joinpath("switches"))
    infrable.switches(repr=True, _out=data.joinpath("switches_repr"))
    infrable.switches(
        format="{name},{foo},{init}", _out=data.joinpath("switches_format")
    )


def setup_module():
    infrable.switch.env.dev()
    # gen_test_data()
    init.init()


def test_switches():
    assert infrable.switches(_tty_out=False) == data.joinpath("switches").read_text()
    assert "env = Switch(options={'" in data.joinpath("switches_repr").read_text()
    assert (
        infrable.switches(format="{name},,{init}", _tty_out=False)
        == data.joinpath("switches_format").read_text()
    )



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional

import yaml
from jinja2 import StrictUndefined
from jinja2 import Template as JinjaTemplate

from infrable import errors


def _jinja_render(content: str, *args: dict[str, Any], **kwargs: Any) -> str:
    return JinjaTemplate(content, undefined=StrictUndefined).render(*args, **kwargs)


@dataclass(frozen=True)
class Dest:
    """Destination for the rendered template, with optional context."""

    loc: str
    chown: Optional[str] = None
    chmod: Optional[str] = None
    ctx: dict = field(default_factory=dict)

    @classmethod
    def parse(cls, data: str | dict):
        """Parse a destination from the given string or dictionary."""

        if isinstance(data, str):
            return cls(loc=data)

        return cls(**data)


@dataclass(frozen=True)
class Metadata:
    """Metadata defined in each template."""

    src: str
    dest: list[Dest]
    chown: Optional[str] = None
    chmod: Optional[str] = None
    skip: bool = False

    _raw: str = ""
    _rendered: str = ""

    @classmethod
    def parse(cls, path: Path, *args: dict[str, Any], **kwargs: Any):
        """Parse metadata from the given file content."""

        reading = False
        raw = ""
        rendered = ""
        for line in path.read_text().splitlines():
            if line == "# ---":
                if reading:
                    break
                else:
                    reading = True
                    continue
            if reading:
                raw += f"{line}\n"

        raw = raw.strip()

        kwargs["_template"] = {"src": str(path)}
        rendered = _jinja_render(raw, *args, **kwargs)
        yamllines = [line.removeprefix("# ") for line in rendered.splitlines()]
        if not yamllines:
            raise errors.MissingMetadataError(src=path)

        meta: dict = yaml.safe_load("\n".join(yamllines))
        meta_dest: str | dict | list[str | dict] | None = meta.pop("dest")

        if not meta_dest:
            raise errors.MissingDestinationError(src=meta["src"])

        dest = []
        if isinstance(meta_dest, list):
            dest = [Dest.parse(d) for d in meta_dest]
        else:
            dest = [Dest.parse(meta_dest)]

        return cls(dest=dest, _raw=raw, _rendered=rendered, **meta)

    def clean_header(self, dest_loc: str) -> str:
        dest = next(d for d in self.dest if d.loc == dest_loc)

        # We don't need "skip" in the final header
        dict_: dict[str, Any] = {"src": self.src, "dest": dest.loc}

        # We need chown and chmod in the final header to diff file permissions
        chown = dest.chown or self.chown
        if chown:
            dict_["chown"] = str(chown)
        chmod = dest.chmod or self.chmod
        if chmod:
            dict_["chmod"] = str(chmod)

        header = yaml.safe_dump(dict_, default_flow_style=False)
        header = "\n".join(f"# {l}" for l in header.splitlines())
        return header


def render(path: Path, dest_loc: str, *args: dict[str, Any], **kwargs: Any) -> str:
    """Render the template with the given context, for a single destination."""

    meta = Metadata.parse(path, *args)
    header = meta.clean_header(dest_loc)
    tmpl = path.read_text().replace(meta._raw, header)
    content = _jinja_render(tmpl, *args, **kwargs)

    return f"{content}\n"  # all config, specially sudoers requires a trailing newline



from sh import infrable

from infrable import __version__, init


def setup_module():
    init.init()


def test_custom_module():
    assert "[WORKFLOW] Provision Ubuntu host" in infrable.cloud(
        "provision-ubuntu-host", "--help"
    )



from pathlib import Path

import typer

from infrable import files

app = typer.Typer(no_args_is_help=True)


@app.command(help=files.affected_hosts.__doc__)
def affected_hosts(only: list[str] = typer.Option(None)):
    for host in files.affected_hosts(only=only):
        print(host)


@app.command(help=files.deploy.__doc__)
def deploy(
    path: Path = typer.Argument(None),
    only: list[str] = typer.Option(None),
    yes: bool = False,
):
    files.deploy(path, only=only, yes=yes)


@app.command(help=files.recover.__doc__)
def recover(path: Path = typer.Argument(None)):
    files.recover(path)


@app.command(help=files.gen.__doc__)
def gen(
    path: Path = typer.Argument(None),
    only: list[str] = typer.Option(None),
):
    files.gen(path, only=only)


app.command()(files.backup)
app.command()(files.pull)
app.command()(files.diff)
app.command()(files.push)


@app.command(help=files.revert.__doc__)
def revert(path: Path = typer.Argument(None)):
    files.revert(path)



import typer
from click import Choice

app = typer.Typer(no_args_is_help=True)

from infrable import IS_PROJECT_DIR, __version__, init


@app.command()
def version():
    """Print the version."""
    print(f"infrable {__version__}")


@app.command(name="init")
def init_():
    """Bootstrap a dummy project."""
    init.init()


if IS_PROJECT_DIR:
    from infrable import infra
    from infrable.cli import files, remote, switch

    @app.command()
    def hosts(format: str | None = None, repr: bool = False):
        """List all hosts in the infrastructure."""

        for name, host in infra.hosts.items():
            if format:
                print(host.format(name, format=format))
            elif repr:
                print(f"{name} = {host.__repr__()}")
            else:
                print(f"{name} = {host}")

    @app.command()
    def services(format: str | None = None, repr: bool = False):
        """List all services in the infrastructure."""

        for name, service in infra.services.items():
            if format:
                print(service.format(name, format=format))
            elif repr:
                print(f"{name} = {service.__repr__()}")
            else:
                print(f"{name} = {service}")

    @app.command()
    def switches(format: str | None = None, repr: bool = False):
        """List all switches in the infrastructure."""

        for name, switch in infra.switches.items():
            if format:
                print(switch.format(name, format=format))
            elif repr:
                print(f"{name} = {switch.__repr__()}")
            else:
                print(f"{name} = {switch}")

    app.add_typer(files.app, name="files", help="Manage files.")
    app.add_typer(switch.app, name="switch", help="Manage switches.")
    app.add_typer(remote.app, name="remote", help="Execute remote commands.")

    for name, ext in infra.typers.items():
        app.add_typer(ext, name=name.replace("_", "-"))

    if sw := infra.switches.get("env"):
        # Alias for 'infrable switch env'
        def env(value: str = typer.Argument(..., click_type=Choice(list(sw.options)))):
            """Alias for `infrable switch env`."""
            sw.set(value)

        app.command()(env)



from sh import infrable

from infrable import __version__, init


def setup_module():
    init.init()


def test_workflows():
    assert "[WORKFLOW] Deploy nginx files" in infrable.deploy.nginx("--help")
